package org.maltparser.core.graph;


import java.util.HashMap;
import java.util.SortedMap;
import java.util.SortedSet;

import org.maltparser.core.dataformat.ColumnDescription;
import org.maltparser.core.dataformat.DataFormatInstance;
import org.maltparser.core.exception.MaltChainedException;
import org.maltparser.core.graph.arc.ArcLabels;
import org.maltparser.core.graph.Node;
import org.maltparser.core.sentence.ColumnInstance;
import org.maltparser.core.sentence.Sentence;
import org.maltparser.core.symbol.SymbolTable;
import org.maltparser.transform.PseudoProjectivity;

/**
 *  
 *
 * @author Johan Hall
 * @since 1.0
**/
public class DependencyGraph {
	protected Node[] nodes;
	protected int capacity;
	protected int size;
	protected Sentence sentence;
	protected RootLabels rootLabels;
	protected HashMap<SymbolTable,SymbolTable> outputColumnMap; // to handle DEPREL_PPROJ
	protected boolean singleheaded;
	protected int numberOfComponents;

	public DependencyGraph(Sentence sentence, String rootLabel) throws MaltChainedException {
		setSize(0);
		setSentence(sentence);
		initNodes(rootLabel);
		
		outputColumnMap = new HashMap<SymbolTable,SymbolTable>();
		for (ColumnDescription column : sentence.getDataFormatInstance()) {
			if (column.getCategory() == ColumnDescription.OUTPUT) {
				outputColumnMap.put(column.getSymbolTable(), column.getSymbolTable());
			}
		}
	}
	
	public void addArc(Node head, Node dependent, ArcLabels arcLabelCodes) throws MaltChainedException {
		if (dependent.getIndex().intValue() < 0 && dependent.getIndex().intValue() > getSize()) {
			throw new GraphException("The depedent node is out of bound when adding an arc. ");
		} else if (head.getIndex().intValue() < 0 && head.getIndex().intValue() > getSize()) {
			throw new GraphException("The head node is out of bound when adding an arc. ");
		} 
		Node hc = nodes[head.getIndex()].findComponent();
		Node dc = nodes[dependent.getIndex()].findComponent();
		if (hc != dc) {
			link(hc, dc);
			numberOfComponents--;		
		}
		nodes[dependent.getIndex()].addHead(nodes[head.getIndex()], arcLabelCodes);
		if (singleheaded == true) {
			singleheaded = nodes[dependent.getIndex()].isSingleHeaded();
		}
	}
	
	public DataFormatInstance getDataFormatInstance() {
		return sentence.getDataFormatInstance();
	}
	
	public Node getRoot() throws MaltChainedException {
		if (nodes[0].isRoot()) {
			return nodes[0];
		} else {
			throw new GraphException("The dependency graph does not have any root. ");
		}	
	}
	
	public boolean hasArc(int headindex, int depindex) {
		return (nodes[depindex].getHead(headindex) != null);
	}
	
	public ArcLabels getArcLabels(int headindex, int depindex) {
		return nodes[depindex].getArcLabels(headindex);
	}
	
	public Integer getLabelCode(int headindex, int depindex, SymbolTable table) {
		return nodes[depindex].getLabelCode(headindex, table);
	}
	
	public Node getNode(int index) {
		if (index >= 0 && index <= size) {
			return nodes[index];
		}
		return null;
	}
	
	public Node getTokenNode(int index) {
		if (index > 0 && index <= size && !nodes[index].isRoot()) {
			return nodes[index];
		}
		return null;
	}

	public RootLabels getRootLabels() throws MaltChainedException {
		if (nodes[0].isRoot()) {
			return rootLabels;
		} else {
			throw new GraphException("The dependency graph does not have any root. ");
		}
	}
	
	protected Node link(Node x, Node y) {
		if (x.getRank() > y.getRank()) {
			y.setComponent(x);
		} else {
			x.setComponent(y);
			if (x.getRank() == y.getRank()) {
				y.setRank(y.getRank()+1);
			}
			return y;
		}
		return x;
	}
	
	
	/*link (int i, int j) {
		  if (rank[i] > rank[j]) {
			root[j] = i;
		  }
		  else {
		    root[i] = j;
		    if (rank[i] == rank[j]) {
		      rank[j] = j+1;
	            }
	 	  }
		}*/
	public boolean isSingleHeaded() {
		return singleheaded;
	}
	

	
	public boolean isConnected() {
		return (numberOfComponents == 1);
	}
	
	public boolean isTree() {
		return isConnected() && singleheaded;
	}
	

	

	
	public void linkAllTokensToRoot() throws MaltChainedException {
		clear();
		for (int i = 1; i <= size; i++) {
			nodes[i].addHead(getRoot());
		}
	}

	public void linkAllTreesToRoot() throws MaltChainedException {
		for (int i = 1; i <= size; i++) {
			if (!nodes[i].hasHead()) {
				nodes[i].addHead(getRoot());
			}
		}
	}
	
	public void clear() {
		nodes[0].clear();
		if (size + 1 < capacity) {
			nodes[size].setSuccessor(nodes[size + 1]);
		}
		for (int i = 1; i <= size; i++) {
			nodes[i].clear();
		}
		size = sentence.getSize();
		nodes[size].setSuccessor(null);
		
		singleheaded = true;
		numberOfComponents = size + 1;
	}

	public boolean linkedLeftOf(int right, int left) {
		if (nodes[right].hasHead() && nodes[right].getHead().getIndex().intValue() < left) {
			return true;
		}
		if (nodes[right].hasLeftDependent() && nodes[right].getLeftmostDependent().getIndex().intValue() < left) {
			return true;
		}
		return false;
	}
	
	public boolean hasLabeledDependency(Node dependent, SymbolTable table) {
		return (!dependent.isRoot() && dependent.getLabelCode(table) != null);
	}

	public boolean hasLabeledDependency(Node dependent) {
		return (!dependent.isRoot() && dependent.hasLabelCodes());
	}
	
	public int getCapacity() {
		return capacity;
	}
	
	public int getSize() {
		return size;
	}	
	
	public Sentence getSentence() {
		return sentence;
	}
	
	public void createFromGoldStandard()  throws MaltChainedException {
		clear();
		ColumnDescription headColumn = getDataFormatInstance().getHeadColumnDescription();
		SortedMap<String,ColumnDescription> outputColumns = getDataFormatInstance().getOutputColumnDescriptions();

		if (headColumn == null) {
			return;
		}
		if (outputColumns.size() == 0) {
			for (int i = 1; i <= size; i++) {
				if (!nodes[i].isRoot()) {
					int head = nodes[i].getToken().getFirstHeadColumnInstance().getIntegerValue();
					if (head == 0) {
						nodes[i].addHead(getRoot());
					} else {
						nodes[i].addHead(nodes[head]);
					}
				}
			}
		} else {
			for (int i = 1; i <= size; i++) {
				if (!nodes[i].isRoot()) {
					int head = nodes[i].getToken().getFirstHeadColumnInstance().getIntegerValue();
					SortedSet<ColumnInstance> outputColumnAttributes = nodes[i].getToken().getOutputColumnInstances();
					ArcLabels arcLabels = new ArcLabels();
					for (ColumnInstance outputColumnAttribute : outputColumnAttributes) {
						arcLabels.put(outputColumnAttribute.getColumnDescription().getSymbolTable(), outputColumnAttribute.getIntegerValue());
					}
					if (head == 0) {
						nodes[i].addHead(getRoot(), arcLabels);
					} else {
						nodes[i].addHead(nodes[head], arcLabels);
					}
				}
			}
		}
	}
	
	public void projectivize(String markingStrategy, String coveredRoot, String liftingOrder) throws MaltChainedException {
		PseudoProjectivity pproj = new PseudoProjectivity(markingStrategy, coveredRoot, liftingOrder, sentence.getLogger(), sentence.getDataFormatInstance().getSymbolTables().getSymbolTable("DEPREL"));
		pproj.projectivize(this);
	}
	
	public void deprojectivize(String markingStrategy) throws MaltChainedException {
		PseudoProjectivity pproj = new PseudoProjectivity(markingStrategy, "none", "shortest", sentence.getLogger(), sentence.getDataFormatInstance().getSymbolTables().getSymbolTable("DEPREL"));
		pproj.deprojectivize(this);
	}
	
	public String toTab() throws MaltChainedException {
		StringBuilder sb = new StringBuilder();
		for (int i = 1; i <= size; i++) {
			int j=0;
			for (ColumnDescription column : sentence.getDataFormatInstance()) {
				if (j != 0) {
					sb.append('\t');
				}
				if (column.getType() == ColumnDescription.IGNORE) {
					sb.append(column.getDefaultOutput());
				} else if (column.getCategory() == ColumnDescription.INPUT || column.getCategory() == ColumnDescription.ECHO) {
					sb.append(nodes[i].getToken().getColumnInstance(column).getStringValue());
				} else if (column.getCategory() == ColumnDescription.OUTPUT || column.getCategory() == ColumnDescription.HEAD) {
					int k = 0;
					if (nodes[i].getHeads() == null) {continue;}
					for (Node head :  nodes[i].getHeads()) {
						if (k != 0) {
							sb.append('|');
						}
						if (column.getCategory() == ColumnDescription.HEAD) {
							sb.append(head.getIndex().intValue());
						} else if (column.getCategory() == ColumnDescription.OUTPUT) {
							SymbolTable table = outputColumnMap.get(column.getSymbolTable());
							Integer labelCode = nodes[i].getLabelCode(head, table);
							if (labelCode == null) {
								sb.append(rootLabels.getRootLabel(table));
							} else if (labelCode.equals(rootLabels.getRootLabelCode(column.getSymbolTable()))) {
								sb.append(rootLabels.getRootLabel(table));
							} else {
								sb.append(table.getSymbolCodeToString(labelCode));
							}
						}
					}
				}
				j++;
			}
			sb.append('\n');
		}
		sb.append('\n');
		return sb.toString();
	}
	
	public void changeOutputColumnSymbolTable(SymbolTable columnTable, SymbolTable table) {
		outputColumnMap.put(columnTable, table);
	}
	
	protected void initNodes(String rootLabel) throws MaltChainedException {
		if (sentence != null) {
			nodes = new Node[capacity];
			nodes[0] = new Node(0,true,false); 
			rootLabels = new RootLabels(rootLabel, sentence.getDataFormatInstance().getOutputColumnDescriptions());
			for (int i = 1; i < capacity; i++) {
				nodes[i] = new Node(i, false, false, sentence.getToken(i));
			}

			for (int i = 1; i < capacity - 1; i++) {
				nodes[i].setSuccessor(nodes[i + 1]);
			}
			for (int i = 2; i < capacity; i++) {
				nodes[i].setPredecessor(nodes[i - 1]);
			}
		}
	}

	protected void setCapacity(int capacity) {
		this.capacity = capacity;
	}

	protected void setSize(int size) {
		this.size = size;
	}
	
	protected void setSentence(Sentence sentence) {
		this.sentence = sentence;
		if (sentence != null) {
			setCapacity(sentence.getCapacity());
		}
	}
}
