package org.maltparser.core.graph;

import java.util.NoSuchElementException;
import java.util.SortedSet;
import java.util.TreeSet;

import org.maltparser.core.exception.MaltChainedException;
import org.maltparser.core.graph.arc.ArcLabels;
import org.maltparser.core.graph.arc.HeadArcs;
import org.maltparser.core.graph.arc.MultipleHeadArcs;
import org.maltparser.core.graph.arc.SingleHeadArc;
import org.maltparser.core.graph.Node;
import org.maltparser.core.sentence.Token;
import org.maltparser.core.symbol.SymbolTable;
import org.maltparser.parser.guide.feature.Featurable;

/**
 *  
 *
 * @author Johan Hall
 * @since 1.0
**/
public class Node implements Comparable<Node>, Featurable {
	/**
	 * the node index
	 */
	protected Integer index;
	/**
	 * the previous node in the linear precedence
	 */
	protected Node predecessor;
	/**
	 * the next node in the linear precedence
	 */
	protected Node successor;
	/**
	 * a sorted set of left dependent nodes (sorted according to the node index) 
	 */
	protected SortedSet<Node> leftDependents;
	/**
	 * a sorted set of right dependent nodes (sorted according to the node index) 
	 */
	protected SortedSet<Node> rightDependents;
	/**
	 * the head arcs that contains one or more labeled arcs (if <code>singleHeadedConstraint==true</code>, it 
	 * contains only one arc per node) 
	 */
	protected HeadArcs arcs;
	/**
	 * a reference to the token in the sentence
	 */
	protected Token token;
	/**
	 * a reference to a node where the node is part of a component. If the node is unconnected it will reference to it self. 
	 */
	protected Node component;
	protected int rank;
	/**
	 * <code>true</code> if the node is a root, otherwise <code>false</code>
	 */
	protected boolean root;
	protected boolean singleHeadedConstraint;
	
	public Node(int index, boolean isRoot, boolean isSingleHeaded) {
		this(index, isRoot, isSingleHeaded, null);
	}
	
	public Node(int index, boolean isRoot, boolean isSingleHeaded, Token token) {
		setIndex(index);
		setPredecessor(null);
		setSuccessor(null);
		
		setRoot(isRoot);
		setSingleHeadedConstraint(isSingleHeaded);
		
		component = this;
		rank = 0;
		
		this.leftDependents = new TreeSet<Node>();
		this.rightDependents = new TreeSet<Node>();
		initArcs();
		setToken(token);
	}
	
	public void addHead(Node head) throws MaltChainedException {
		this.addHead(head, null);
	}
	
	public void addHead(Node head, ArcLabels arcLabelCodes) throws MaltChainedException {
		if (!isRoot()) {
			if (isSingleHeadedConstraint() && arcs.hasHead()) {
				arcs.getHead().removeDependent(this);
				//throw new GraphException("It is not allowed to give a singleheaded node more than one head.");
			}
			arcs.addHead(head, arcLabelCodes);
			head.addDependent(this);
		} else {
			throw new GraphException("It is not allowed to assign a head to the root node. ");
		}
	}
	
	public int getRank() {
		return rank;
	}
	
	public void setRank(int r) {
		this.rank = r;
	}
	
	public Node findComponent() {
		return findComponent(this);
	}
	
	private Node findComponent(Node x) {
		if (x != x.getComponent()) {
			x.setComponent(findComponent(x.getComponent()));
		}
		return x.getComponent();
	}
	
	public Node getComponent() {
		return component;
	}
	
	public void setComponent(Node x) {
		this.component = x;
	}
	
	public boolean isSingleHeaded() {
		return arcs.isSingleHeaded();
	}
	
	public Node getHead(int index) {
		if (!isRoot()) {
			return arcs.getHead(index);
		}
		return null;
	}
	
	public Integer getLabelCode(Node head, SymbolTable table) {
		if (!isRoot()) {
			return arcs.getLabelCode(head, table);
		}
		return null;
	}
	
	public Integer getLabelCode(SymbolTable table) {
		if (!isRoot()) {
			return arcs.getFirstHeadLabelCode(table);
		}
		return null;
	}
	
	public ArcLabels getLabelCode() {
		if (!isRoot()) {
			return arcs.getFirstHeadArcLabels();
		}
		return null;
	}
	
	public boolean hasLabelCodes() {
		return arcs.hasLabelCodes();
	}
	
	public Node getHead() {
		if (!isRoot()) {
			return arcs.getHead();
		}
		return null;
	}
	
	public SortedSet<Node> getHeads() {
		if (!isRoot()) {
			arcs.getHeads();
		}
		return null;
	}
	
	public Integer getLabelCode(int headindex, SymbolTable table) {
		if (!isRoot()) {
			return arcs.getLabelCode(headindex, table);
		}
		return null;
	}

	public ArcLabels getArcLabels(int headindex) {
		if (!isRoot()) {
			return arcs.getArcLabels(headindex);
		}
		return null;
	}

	public Integer getHeadLabelCode(SymbolTable table) {
		if (!isRoot()) {
			arcs.getFirstHeadLabelCode(table);
		}
		return null;
	}
	
	protected void addDependent(Node dependent) throws MaltChainedException {
		if (compareTo(dependent) > 0) {
			leftDependents.add(dependent);
		} else if (compareTo(dependent) < 0) {
			rightDependents.add(dependent);
		} else {
			throw new GraphException("It is not allowed to link a node to itself. ");
		}
	}
	protected void removeDependent(Node dependent) {
		if (compareTo(dependent) > 0) {
			leftDependents.remove(dependent);
		} else if (compareTo(dependent) < 0) {
			rightDependents.remove(dependent);
		}
	}
	
	public void clear() {
		leftDependents.clear();
		rightDependents.clear();

		component = this;
		rank = 0;
		
		if (!isRoot()) {
			arcs.clear();
		}
	}
	
	public boolean hasAncestorInside(int left, int right) {
		Node tmp = this;
		
		if (tmp.getHead() != null) {
			tmp = tmp.getHead();
			if (tmp.getIndex().intValue() >= left && tmp.getIndex().intValue() <= right) {
				return true;
			}
		}
		/* while (tmp.getHead() != null) {
			tmp = tmp.getHead();
			if (tmp.getIndex().intValue() >= left && tmp.getIndex().intValue() <= right) {
				return true;
			}
		}*/
		return false;
	}
	
	public boolean isRoot() {
		return root;
	}
	
	public boolean isSingleHeadedConstraint() {
		return singleHeadedConstraint;
	}

	public Integer getIndex() {
		return index;
	}
	
	public Token getToken() {
		return token;
	}
	
	public boolean hasHead() {
		if (!isRoot()) {
			return arcs.hasHead();
		}
		return false;
	}
	
	/**
	 * Returns the predecessor node in the input queue
	 * 
	 * @return the predecessor node in the input queue
	 */
	public Node getPredecessor() {
		return predecessor;
	}

	/**
	 * Returns the successor node in the input queue
	 * 
	 * @return the successor node in the input queue
	 */
	public Node getSuccessor() {
		return successor;
	}

	/**
	 * Returns the left sibling if it exists, otherwise <code>null</code>
	 * 
	 * @return the left sibling if it exists, otherwise <code>null</code>
	 */
	public Node getLeftSibling() {
		if (getHead() == null) {
			return null;
		}
		SortedSet<Node> union = new TreeSet<Node>(getHead().getLeftDependents());
		union.addAll(getHead().getRightDependents());

		try {
			SortedSet<Node> headSet = union.headSet(this);
			return headSet.last();
		} catch (NoSuchElementException e) {
			return null;
		}
	}

	/**
	 * Returns the left sibling at the same side of head as the node it self. If not found <code>null</code is returned
	 * 
	 * @return the left sibling at the same side of head as the node it self. If not found <code>null</code is returned
	 */
	public Node getSameSideLeftSibling() {
		if (getHead() == null) {
			return null;
		} else if (this.getIndex().intValue() < getHead().getIndex().intValue()) {
			try {
				return getHead().getLeftDependents().headSet(this).last();
			} catch (NoSuchElementException e) {
				return null;
			}
		} else if (this.getIndex().intValue() > getHead().getIndex().intValue()) {
			try {
				return getHead().getRightDependents().headSet(this).last();
			} catch (NoSuchElementException e) {
				return null;
			}
		}
		return null;
	}
	
	/**
	 * Returns the right sibling if it exists, otherwise <code>null</code>
	 * 
	 * @return the right sibling if it exists, otherwise <code>null</code>
	 */
	public Node getRightSibling() {
		if (getHead() == null) {
			return null;
		}
		SortedSet<Node> union = new TreeSet<Node>(getHead().getLeftDependents());
		union.addAll(getHead().getRightDependents());
		SortedSet<Node> tailSet = union.tailSet(this);
		
		if (tailSet.size() <= 1)  {
			return null;
		}
		return tailSet.toArray(new Node[tailSet.size()])[1];
	}

	/**
	 * Returns the right sibling at the same side of head as the node it self. If not found <code>null</code is returned
	 * 
	 * @return the right sibling at the same side of head as the node it self. If not found <code>null</code is returned
	 */
	public Node getSameSideRightSibling() {
		if (getHead() == null) {
			return null;
		} else if (this.getIndex().intValue() < getHead().getIndex().intValue()) {
			SortedSet<Node> tailSet = getHead().getLeftDependents().tailSet(this);
			if (tailSet.size() <= 1)  {
				return null;
			}
			return tailSet.toArray(new Node[tailSet.size()])[1];
		} else if (this.getIndex().intValue() > getHead().getIndex().intValue()) {
			SortedSet<Node> tailSet = getHead().getRightDependents().tailSet(this);
			if (tailSet.size() <= 1)  {
				return null;
			}
			return tailSet.toArray(new Node[tailSet.size()])[1];
		}
		return null;
	}
	
	/**
	 * Returns <code>true</code> if the node has one or more left dependents, otherwise <code>false</code>.
	 * 
	 * @return <code>true</code> if the node has one or more left dependents, otherwise <code>false</code>.
	 */
	public boolean hasLeftDependent() {
		return !leftDependents.isEmpty();
	}

	/**
	 * Returns the left most dependent. If the node does not have any left dependent(s) <code>null</code> is returned.
	 * 
	 * @return the left most dependent. If the node does not have any left dependent(s) <code>null</code> is returned.
	 */
	public Node getLeftmostDependent() {
		try {
			return leftDependents.first();
		} catch (NoSuchElementException e) {
			return null;
		}
	}

	/**
	 * Returns the closest left dependent to the node it self, if not found <code>null</code> is returned.
	 * 
	 * @return the closest left dependent to the node it self, if not found <code>null</code> is returned.
	 */
	public Node getClosestLeftDependent() {
		try {
			return leftDependents.last();
		} catch (NoSuchElementException e) {
			return null;
		}
	}
	
	/**
	 * Returns a sorted set of left dependents 
	 * 
	 * @return a sorted set of left dependents 
	 */
	public SortedSet<Node> getLeftDependents() {
		return leftDependents;
	}
	
	/**
	 * Returns the left dependent at the position <code>index</code>, where <code>index==0</code> equals the left most dependent.
	 * 
	 * @param index the index
	 * @return the left dependent at the position <code>index</code>, where <code>index==0</code> equals the left most dependent
	 */
	public Node getLeftDependent(int index) {
		int size = leftDependents.size();
		if (0 <= index && index < size) {
			return leftDependents.toArray(new Node[size])[index];
		}
		return null;
	}
	
	/**
	 * Return the number of left dependents
	 * 
	 * @return the number of left dependents
	 */
	public int getLeftDependentCount() {
		return leftDependents.size();
	}
	
	/**
	 * Returns <code>true</code> if the node has one or more right dependents, otherwise <code>false</code>.
	 * 
	 * @return <code>true</code> if the node has one or more right dependents, otherwise <code>false</code>.
	 */
	public boolean hasRightDependent() {
		return !rightDependents.isEmpty();
	}
	
	/**
	 * Returns the right most dependent. If the node does not have any right dependent(s) <code>null</code> is returned.
	 * 
	 * @return the right most dependent. If the node does not have any right dependent(s) <code>null</code> is returned.
	 */
	public Node getRightmostDependent() {
		try {
			return rightDependents.last();
		} catch (NoSuchElementException e) {
			return null;
		}
	}

	/**
	 * Returns the closest right dependent to the node it self, if not found <code>null</code> is returned.
	 * 
	 * @return the closest right dependent to the node it self, if not found <code>null</code> is returned.
	 */
	public Node getClosestRightDependent() {
		try {
			return rightDependents.first();
		} catch (NoSuchElementException e) {
			return null;
		}
	}

	/**
	 * Returns a sorted set of right dependent nodes
	 * 
	 * @return a sorted set of right dependent nodes
	 */
	public SortedSet<Node> getRightDependents() {
		return rightDependents;
	}
	
	/**
	 * Returns the right dependent at the position <code>index</code>, where <code>index==0</code> equals the right most dependent.
	 * 
	 * @param index the index
	 * @return the right dependent at the position <code>index</code>, where <code>index==0</code> equals the right most dependent
	 */
	public Node getRightDependent(int index) {
		int size = rightDependents.size();
		if (index < size) {
			return rightDependents.toArray(new Node[size])[size - 1 - index];
		}
		return null;
	}
	
	/**
	 * Return the number of right dependents
	 * 
	 * @return the number of right dependents
	 */
	public int getRightDependentCount() {
		return rightDependents.size();
	}
	
	
	public Node getLeftmostDescendant() {
		int i;
		Node leftMost = this, child, descendant;
		i = 0;
		while ((child = getLeftDependent(i)) != null) {
			descendant = child.getLeftmostDescendant();
			if (descendant.compareTo(leftMost) < 0) {
				leftMost = descendant;
			}
			i++;
		}
		i = 0;
		while ((child = getRightDependent(i)) != null) {
			descendant = child.getLeftmostDescendant();
			if (descendant.compareTo(leftMost) < 0) {
				leftMost = descendant;
			}
			i++;
		}
		return leftMost;
	}

	public Node getRightmostDescendant() {
		int i;
		Node rightMost = this, child, descendant;
		i = 0;
		while ((child = getLeftDependent(i)) != null) {
			descendant = child.getLeftmostDescendant();
			if (descendant.compareTo(rightMost) > 0) {
				rightMost = descendant;
			}
			i++;
		}
		i = 0;
		while ((child = getRightDependent(i)) != null) {
			descendant = child.getRightmostDescendant();
			if (descendant.compareTo(rightMost) > 0) {
				rightMost = descendant;
			}
			i++;
		}
		return rightMost;
	}
	
	
	public boolean isOperation(String operation) {
		if (operation.equalsIgnoreCase("head")) {
			return true;
		} else if (operation.equalsIgnoreCase("ldep")) {
			return true;
		} else if (operation.equalsIgnoreCase("rdep")) {
			return true;
		} else if (operation.equalsIgnoreCase("rdep2")) { 
			// To emulate the behavior of MaltParser 0.4 (bug)
			return true;
		} else if (operation.equalsIgnoreCase("lsib")) {
			return true;
		} else if (operation.equalsIgnoreCase("rsib")) {
			return true;
		} else if (operation.equalsIgnoreCase("pred")) {
			return true;
		} else if (operation.equalsIgnoreCase("succ")) {
			return true;
		}
		return false;
	}

	public Node getNode(String operation, int index) {
		if (operation.equalsIgnoreCase("head")) {
			return getHead();
		} else if (operation.equalsIgnoreCase("ldep")) {
			return getLeftmostDependent();
		} else if (operation.equalsIgnoreCase("rdep")) {
			return getRightmostDependent();
		} else if (operation.equalsIgnoreCase("rdep2")) { 
			// To emulate the behavior of MaltParser 0.4 (bug)
			if (!this.isRoot()) {
				return getRightmostDependent();
			}
		} else if (operation.equalsIgnoreCase("lsib")) {
			return getSameSideLeftSibling();
		} else if (operation.equalsIgnoreCase("rsib")) {
			return getSameSideRightSibling();
		} else if (operation.equalsIgnoreCase("pred")) {
			return getPredecessor();
		} else if (operation.equalsIgnoreCase("succ")) {
			return getSuccessor();
		}
		return null;
	}
	
	/* Setters */
	protected void setIndex(Integer index) {
		this.index = index;
	}
	
	public void setPredecessor(Node predecessor) {
		this.predecessor = predecessor;
	}
	
	public void setSuccessor(Node successor) {
		this.successor = successor;
	}
	
	protected void setRoot(boolean root) {
		this.root = root;
	}

	protected void setSingleHeadedConstraint(boolean singleHeadedConstraint) {
		this.singleHeadedConstraint = singleHeadedConstraint;
	}

	protected void setToken(Token token) {
		if (isRoot()) {
			this.token = null;
		} else {
			this.token = token;
		}
	}
	
	protected void initArcs() {
		if (isRoot()) {
			arcs = null;
		} else {
			if (singleHeadedConstraint) {
				arcs = new SingleHeadArc();
			} else {
				arcs = new MultipleHeadArcs();
			}
		}
	}
	
	public int compareTo(Node o) {
		return index.compareTo(o.getIndex());
	}
}
