package org.maltparser.parser.guide.feature;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.regex.PatternSyntaxException;

import org.maltparser.core.exception.MaltChainedException;
import org.maltparser.core.symbol.SymbolTable;
import org.maltparser.parser.algorithm.ParsingAlgorithm;


public class SplitFeature extends Feature {
	private Feature feature;
	private SymbolTable table;
	private String separators;
	
	public SplitFeature(ParsingAlgorithm parsingAlgorithm, String spec) throws MaltChainedException {
		super();
		super.initialize(parsingAlgorithm);
		spec = spec.trim();
		int index = spec.lastIndexOf(',');
		if (index == -1) {
			throw new FeatureException("The split feature '"+spec+"' has not the correct format. ");
		}
		if (spec.substring(0,index).trim().startsWith("Split")) {
			throw new FeatureException("Could not have recursive split feature at line '"+spec+"'. ");
		}
		
		this.feature = initParentFeature(spec.substring(0,index).trim());
		if (feature instanceof InputColumnFeature || feature instanceof OutputColumnFeature) {
			table = parsingAlgorithm.getConfiguration().getSymbolTables().addSymbolTable("SPLIT_"+((ColumnFeature)feature).getColumn().getName(),
					((ColumnFeature)feature).getSymbolTable());
		} else {
			throw new FeatureException("Could not use split feature because only column feature is supported. ");
		}
		if (!init(spec.substring(index+1))) { 
			throw new FeatureException("The split feature could not be initialized at line '"+spec+"'. ");
		}
	}

	protected Feature initParentFeature(String spec) throws MaltChainedException {
		spec = spec.trim();
		 int index = spec.indexOf('(');
		 if (index == -1) {
			 throw new FeatureException("There is no start parenthesis at line '"+spec+"' in the feature specification file. ");
		 }
		 if (!spec.endsWith(")")) {
			 throw new FeatureException("There is no end parenthesis at line '"+spec+"' in the feature specification file. ");
		 }
		 Class<?> clazz;
		 try {
			clazz = Class.forName("org.maltparser.parser.guide.feature."+spec.substring(0, index)+"Feature");
		 } catch (ClassNotFoundException e) {
			 throw new FeatureException("The feature class '"+spec.substring(0, index)+"' cannot be found. ", e);
		 }
		 try {
			Class<?>[] argTypes = { org.maltparser.parser.algorithm.ParsingAlgorithm.class, java.lang.String.class };
			Object[] arguments = new Object[2];
			arguments[0] = parsingAlgorithm;
			arguments[1] = spec.substring(index+1,spec.length()-1);
			Constructor<?> constructor = clazz.getConstructor(argTypes);
			return (Feature)constructor.newInstance(arguments);
		} catch (NoSuchMethodException e) {
			throw new FeatureException("The feature '"+clazz.getName()+"' cannot be initialized. ", e);
		} catch (InstantiationException e) {
			throw new FeatureException("The feature '"+clazz.getName()+"' cannot be initialized. ", e);
		} catch (IllegalAccessException e) {
			throw new FeatureException("The feature '"+clazz.getName()+"' cannot be initialized. ", e);
		} catch (InvocationTargetException e) {
			throw new FeatureException("The feature '"+clazz.getName()+"' cannot be initialized. ", e);
		}
	}
	
	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.feature.Feature#init(java.lang.String)
	 */
	public boolean init(String spec) throws MaltChainedException {
		if (spec.trim().length() >= 1) {
			separators = spec.trim();
			return true;
		}
		return false;
	}

	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.feature.Feature#setValue(java.lang.Integer)
	 */
	public void setValue(Integer intValue) throws MaltChainedException {
		//feature.setValue(intValue);
		featureValues.clear();
		featureKnown.clear();
		String value = table.getSymbolCodeToString(intValue);
		if (table.isNullValue(value)) {
			featureValues.put(table.getSymbolStringToCode(value), value);
		} else {
			int code;
			String items[];
			try {
				items = value.split(separators);
			} catch (PatternSyntaxException e) {
				throw new FeatureException("The split feature '"+this.toString()+"' could not split the value using the following separators '"+separators+"'",e);
			}
			for (int i = 0; i < items.length; i++) {
				code = table.addSymbol(items[i]);
				featureValues.put(code, items[i]);
			}
		}
	}
	
	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.feature.Feature#setValue(java.lang.String)
	 */
	public void setValue(String value) throws MaltChainedException {
		//feature.setValue(value);
		featureValues.clear();
		featureKnown.clear();
		if (table.isNullValue(value)) {
			featureValues.put(table.getSymbolStringToCode(value), value);
		} else {
			int code;
			String items[];
			try {
				items = value.split(separators);
			} catch (PatternSyntaxException e) {
				throw new FeatureException("The split feature '"+this.toString()+"' could not split the value using the following separators '"+separators+"'",e);
			}
			for (int i = 0; i < items.length; i++) {
				code = table.addSymbol(items[i]);
				featureValues.put(code, items[i]);
			}
		}
	}

	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.feature.Feature#getStringValue(java.lang.Integer)
	 */
	public String getStringValue(Integer value) throws MaltChainedException {
		return table.getSymbolCodeToString(value);
	}
	
	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.feature.Feature#getIntegerValue(java.lang.String)
	 */
	public Integer getIntegerValue(String value) throws MaltChainedException {
		return table.getSymbolStringToCode(value);
	}

	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.feature.Feature#update()
	 */
	public void update() throws MaltChainedException {
		feature.update();
		featureValues.clear();
		featureKnown.clear();
		String value = feature.getFirstStringValue();
		if (table.isNullValue(value)) {
			featureValues.put(table.getSymbolStringToCode(value), value);
			nullValue = true;
		} else {
			int code;
			String items[];
			try {
				items = value.split(separators);
			} catch (PatternSyntaxException e) {
				throw new FeatureException("The split feature '"+this.toString()+"' could not split the value using the following separators '"+separators+"'",e);
			}
			for (int i = 0; i < items.length; i++) {
				code = table.addSymbol(items[i]);
				featureValues.put(code, items[i]);
				featureKnown.put(code, table.getKnown(items[i]));
			}
			nullValue = false;
		}
	}

	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.feature.Feature#updateCardinality()
	 */
	public void updateCardinality() throws MaltChainedException {
		super.setCardinality(table.getValueCounter()); 
	}
	
	public boolean equals(Object obj) {
		if (!(obj instanceof SplitFeature)) {
			return false;
		}
		if (!obj.toString().equals(this.toString())) {
			return false;
		}
		return true;
	}
	
	public Feature getParentFeature() {
		return feature;
	}
	
	public SymbolTable getSymbolTable() {
		return table;
	}
	
	public String getSeparators() {
		return separators;
	}
	
	
	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.feature.Feature#toString()
	 */
	public String toString() {
		return "Split("+feature.toString()+", "+separators+")";
	}
}

