package org.maltparser.parser.guide.instmodel;

import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Formatter;

import org.maltparser.core.config.SingleMaltConfiguration;
import org.maltparser.core.exception.MaltChainedException;
import org.maltparser.core.graph.DependencyGraph;
import org.maltparser.core.sentence.Sentence;
import org.maltparser.ml.LearningMethod;
import org.maltparser.parser.algorithm.ParsingAlgorithm;
import org.maltparser.parser.algorithm.helper.ParserAction;
import org.maltparser.parser.guide.GuideModel;
import org.maltparser.parser.guide.classtable.ClassTable;
import org.maltparser.parser.guide.feature.Feature;
import org.maltparser.parser.guide.feature.FeatureVector;
import org.maltparser.parser.guide.predmodel.ModelException;

/**

@author Johan Hall
@since 1.0
*/
public class AtomicModel implements InstanceModel {
	private GuideModel parent;
	private String modelName;
	private FeatureVector features;
	private ClassTable classCodeTable;
	private Integer index;
	private Integer frequency = 0;
	private LearningMethod method;

	
	/**
	 * Constructs an atomic model.
	 * 
	 * @param index the index of the atomic model (-1..n), where -1 is special value (used by a single model 
	 * or the master divide model) and n is number of divide models.
	 * @param features the feature vector used by the atomic model.
	 * @param classCodeTable the class code table used by the atomic model.
	 * @param parent the parent guide model.
	 * @throws MaltChainedException
	 */
	public AtomicModel(int index, FeatureVector features, ClassTable classCodeTable, GuideModel parent) throws MaltChainedException {
		setParent(parent);
		setIndex(index);
		if (index == -1) {
			setModelName(parent.getModelName());
		} else {
			setModelName(parent.getModelName()+"."+new Formatter().format("%03d", index));
		}
		setFeatures(features);
		setClassCodeTable(classCodeTable);
		setFrequency(0);
		initMethod();
		if (getParserMode() == SingleMaltConfiguration.LEARN && index == -1) {
			try {
				getConfiguration().getInfoFileWriter().write(method.toString());
				getConfiguration().getInfoFileWriter().flush();
			} catch (IOException e) {
				throw new ModelException("Could not write learner settings to the information file. ", e);
			}
		}
	}
	
	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.model.Model#addInstance(org.maltparser.parser.algorithm.helper.ParserAction)
	 */
	public void addInstance(ParserAction parserAction) throws MaltChainedException {
		try {
			method.addInstance(classCodeTable, features);
		} catch (NullPointerException e) {
			throw new ModelException("The learner cannot be found. ", e);
		}
	}

	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.model.Model#finishedAllData()
	 */
	public void noMoreInstances() throws MaltChainedException {
		try {
			method.noMoreInstances();
		} catch (NullPointerException e) {
			throw new ModelException("The learner cannot be found. ", e);
		}
	}
	
	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.model.Model#finishedSentence(org.maltparser.core.sentence.Sentence, org.maltparser.core.graph.DependencyGraph)
	 */
	public void finalizeSentence(Sentence sentence, DependencyGraph dependencyGraph) throws MaltChainedException {
		try {
			method.finalizeSentence(sentence, dependencyGraph);
		} catch (NullPointerException e) {
			throw new ModelException("The learner cannot be found. ", e);
		}
	}

	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.model.Model#predict()
	 */
	public boolean predict() throws MaltChainedException {
		try {
			if (getParserMode() != SingleMaltConfiguration.PARSE) {
				throw new ModelException("Can only predict during parsing. ");
			}
			return method.predict(features, classCodeTable.getEmptyKBestList());
		} catch (NullPointerException e) {
			throw new ModelException("The learner cannot be found. ", e);
		}
	}


	public void finalize() throws MaltChainedException {
		try {
			if (method != null) {
				method.finalize();
				method = null;
			}
			classCodeTable = null;
			features = null;
			parent = null;
		} catch (NullPointerException e) {
			throw new ModelException("The learner cannot be found. ", e);
		}
	}
	
	/**
	 * Moves all instance from this atomic model into the destination atomic model and add the divide feature.
	 * This method is used by the feature divide model to sum up all model below a certain threshold.
	 * 
	 * @param model the destination atomic model 
	 * @param divideFeature the divide feature
	 * @param divideFeatureIndexVector the divide feature index vector
	 * @throws MaltChainedException
	 */
	public void moveAllInstances(AtomicModel model, Feature divideFeature, ArrayList<Integer> divideFeatureIndexVector) throws MaltChainedException {
		if (method == null) {
			throw new ModelException("The learner cannot be found. ");
		} else if (model == null) {
			throw new ModelException("The guide model cannot be found. ");
		} else if (divideFeature == null) {
			throw new ModelException("The divide feature cannot be found. ");
		} else if (divideFeatureIndexVector == null) {
			throw new ModelException("The divide feature index vector cannot be found. ");
		}
		divideFeature.setValue(index);
		method.moveAllInstances(model.getMethod(), divideFeature, divideFeatureIndexVector);
		method.finalize();
		method = null;
	}
	
	/**
	 * Invokes the train() of the learning method 
	 * 
	 * @throws MaltChainedException
	 */
	public void train() throws MaltChainedException {
		try {
			method.train(features);
			method.finalize();
			method = null;
		} catch (NullPointerException e) {	
			throw new ModelException("The learner cannot be found. ", e);
		}
	}
	
	/**
	 * Initialize the learning method according to the option --learner-method.
	 * 
	 * @throws MaltChainedException
	 */
	public void initMethod() throws MaltChainedException {
		Class<?> clazz = (Class<?>)getConfiguration().getOptionValue("guide", "learner");
		if (clazz == org.maltparser.ml.libsvm.Libsvm.class && (Boolean)getConfiguration().getOptionValue("malt0.4", "behavior") == true) {
			try {
				clazz = Class.forName("org.maltparser.ml.libsvm.malt04.LibsvmMalt04");
			} catch (ClassNotFoundException e) {
				throw new ModelException("Could not find the class 'org.maltparser.ml.libsvm.malt04.LibsvmMalt04'. ", e);
			}
		}
		Class<?>[] argTypes = { org.maltparser.parser.guide.instmodel.InstanceModel.class, java.lang.Integer.class };
		Object[] arguments = new Object[2];
		arguments[0] = this;
		if (getParserMode() == SingleMaltConfiguration.PARSE) {
			arguments[1] = LearningMethod.CLASSIFY;
		} else if (getParserMode() == SingleMaltConfiguration.LEARN) {
			arguments[1] = LearningMethod.TRAIN;
		}

		try {	
			Constructor<?> constructor = clazz.getConstructor(argTypes);
			this.method = (LearningMethod)constructor.newInstance(arguments);
		} catch (NoSuchMethodException e) {
			throw new ModelException("The learner class '"+clazz.getName()+"' cannot be initialized. ", e);
		} catch (InstantiationException e) {
			throw new ModelException("The learner class '"+clazz.getName()+"' cannot be initialized. ", e);
		} catch (IllegalAccessException e) {
			throw new ModelException("The learner class '"+clazz.getName()+"' cannot be initialized. ", e);
		} catch (InvocationTargetException e) {
			throw new ModelException("The learner class '"+clazz.getName()+"' cannot be initialized. ", e);
		}
	}
	
	
	
	/**
	 * Returns the parent guide model
	 * 
	 * @return the parent guide model
	 */
	public GuideModel getParent() throws MaltChainedException {
		if (parent == null) {
			throw new ModelException("The atomic model can only be used by a parent model. ");
		}
		return parent;
	}

	/**
	 * Sets the parent guide model
	 * 
	 * @param parent the parent guide model
	 */
	protected void setParent(GuideModel parent) {
		this.parent = parent;
	}


	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.GuideModel#getModelName()
	 */
	public String getModelName() {
		return modelName;
	}

	/**
	 * Sets the name of the atomic model
	 * 
	 * @param modelName the name of the atomic model
	 */
	protected void setModelName(String modelName) {
		this.modelName = modelName;
	}

	/**
	 * Returns the feature vector used by this atomic model
	 * 
	 * @return a feature vector object
	 */
	public FeatureVector getFeatures() {
		return features;
	}

	/**
	 * Sets the feature vector used by the atomic model.
	 * 
	 * @param features a feature vector object
	 */
	protected void setFeatures(FeatureVector features) {
		this.features = features;
	}

	/**
	 * Returns a class table object
	 * 
	 * @return a class table object
	 */
	public ClassTable getClassCodeTable() {
		return classCodeTable;
	}

	/**
	 * Sets the class table 
	 * 
	 * @param classCodeTable a class table object
	 */
	protected void setClassCodeTable(ClassTable classCodeTable) {
		this.classCodeTable = classCodeTable;
	}

	
	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.GuideModel#getParsingAlgorithm()
	 */
	public ParsingAlgorithm getParsingAlgorithm() throws MaltChainedException {
		try {
			return parent.getParsingAlgorithm();
		} catch (NullPointerException e) {
			throw new ModelException("The parent guide model cannot be found. ", e);
		}
	}
	
	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.GuideModel#getConfiguration()
	 */
	public SingleMaltConfiguration getConfiguration() throws MaltChainedException {
		if (parent != null) {
			return parent.getConfiguration();
		}
		return null;
	}


	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.GuideModel#getParserMode()
	 */
	public int getParserMode() throws MaltChainedException {
		if (parent != null) {
			return parent.getParserMode();
		}
		return -1;
	}

	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.GuideModel#getOutputStreamWriter(java.lang.String)
	 */
	public OutputStreamWriter getOutputStreamWriter(String suffix) throws MaltChainedException {
		if (parent != null) {
			if (index == -1) {
				return parent.getOutputStreamWriter(suffix);
			} else {
				return parent.getOutputStreamWriter(new Formatter().format("%03d", index)+"."+suffix);
			}
		}
		return null;
	}
	
	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.GuideModel#getInputStreamReader(java.lang.String)
	 */
	public InputStreamReader getInputStreamReader(String suffix) throws MaltChainedException {
		if (parent != null) {
			if (index == -1) {
				return parent.getInputStreamReader(suffix);
			} else {
				return parent.getInputStreamReader(new Formatter().format("%03d", index)+"."+suffix);
			}
		}
		return null;
	}
	
	/* (non-Javadoc)
	 * @see org.maltparser.parser.guide.GuideModel#getFile(java.lang.String)
	 */
	public File getFile(String suffix) throws MaltChainedException {
		if (parent != null) {
			if (index == -1) {
				return parent.getFile(suffix);
			} else {
				return parent.getFile(new Formatter().format("%03d", index)+"."+suffix);
			}
		}
		return null;
	}
	
	/**
	 * Returns the index of the atomic model
	 * 
	 * @return the index of the atomic model
	 */
	public Integer getIndex() {
		return index;
	}

	/**
	 * Sets the index of the model (-1..n), where -1 is a special value.
	 * 
	 * @param index index value (-1..n) of the atomic model
	 */
	protected void setIndex(Integer index) {
		this.index = index;
	}

	/**
	 * Returns the frequency (number of instances)
	 * 
	 * @return the frequency (number of instances)
	 */
	public Integer getFrequency() {
		return frequency;
	}
	
	/**
	 * Increase the frequency by 1
	 */
	public void increaseFrequency() {
		if (parent instanceof InstanceModel) {
			((InstanceModel)parent).increaseFrequency();
		}
		frequency++;
	}
	
	public void decreaseFrequency() {
		if (parent instanceof InstanceModel) {
			((InstanceModel)parent).decreaseFrequency();
		}
		frequency--;
	}
	/**
	 * Sets the frequency (number of instances)
	 * 
	 * @param frequency (number of instances)
	 */
	protected void setFrequency(Integer frequency) {
		this.frequency = frequency;
	}
	
	/**
	 * Returns a learner object
	 * 
	 * @return a learner object
	 */
	public LearningMethod getMethod() {
		return method;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append(method.toString());
		return sb.toString();
	}
}
